# Copyright (c) 2020 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from spinn_utilities.overrides import overrides
from pacman.model.partitioner_splitters.abstract_splitters import (
    AbstractSplitterCommon)
from pacman.model.placements import Placement
from pacman.utilities.algorithm_utilities.routing_algorithm_utilities import (
    vertex_xy)
from spinn_front_end_common.data import FecDataView
from .live_packet_gather_machine_vertex import LivePacketGatherMachineVertex


class LPGSplitter(AbstractSplitterCommon):
    """
    Splitter for the :py:class:`LivePacketGather` vertex.
    """

    __slots__ = [
        "__m_vertices_by_ethernet",
        "__targeted_lpgs"
    ]

    def __init__(self):
        super(LPGSplitter, self).__init__()

        self.__m_vertices_by_ethernet = dict()
        self.__targeted_lpgs = set()

    def create_vertices(self, system_placements):
        """
        Special way of making LPG machine vertices, where one is placed
        on each Ethernet-enabled chip.

        .. note::
            This adds to system placements.
        """
        machine = FecDataView.get_machine()
        for eth in machine.ethernet_connected_chips:
            lpg_vtx = LivePacketGatherMachineVertex(
                self.governed_app_vertex.params, self.governed_app_vertex)
            self.governed_app_vertex.remember_machine_vertex(lpg_vtx)
            cores = self.__cores(eth.x, eth.y)
            p = cores[system_placements.n_placements_on_chip(eth.x, eth.y)]
            system_placements.add_placement(
                Placement(lpg_vtx, eth.x, eth.y, p))
            self.__m_vertices_by_ethernet[eth.x, eth.y] = lpg_vtx

    def __cores(self, x, y):
        return [p.processor_id
                for p in FecDataView.get_chip_at(x, y).processors
                if not p.is_monitor]

    @overrides(AbstractSplitterCommon.create_machine_vertices)
    def create_machine_vertices(self, chip_counter):
        # Skip here, and do later!  This is a special case...
        pass

    @overrides(AbstractSplitterCommon.get_in_coming_slices)
    def get_in_coming_slices(self):
        # There are none!
        return []

    @overrides(AbstractSplitterCommon.get_out_going_slices)
    def get_out_going_slices(self):
        # There are also none (but this should never be a pre-vertex)
        return []

    @overrides(AbstractSplitterCommon.get_in_coming_vertices)
    def get_in_coming_vertices(self, partition_id):
        return self.governed_app_vertex.machine_vertices

    @overrides(AbstractSplitterCommon.get_source_specific_in_coming_vertices)
    def get_source_specific_in_coming_vertices(
            self, source_vertex, partition_id):
        # Find the nearest placement for the first machine vertex of the source
        m_vertex = next(iter(source_vertex.splitter.get_out_going_vertices(
            partition_id)))
        x, y = vertex_xy(m_vertex)
        chip = FecDataView.get_chip_at(x, y)
        lpg_vertex = self.__m_vertices_by_ethernet[
            chip.nearest_ethernet_x, chip.nearest_ethernet_y]
        for m_vertex in source_vertex.splitter.get_out_going_vertices(
                partition_id):
            self.__targeted_lpgs.add(
                (lpg_vertex, m_vertex, partition_id))
            lpg_vertex.add_incoming_source(m_vertex, partition_id)
        return [(lpg_vertex, [source_vertex])]

    @property
    def targeted_lpgs(self):
        """
        Which LPG machine vertex is targeted by which machine vertex
        and partition.

        :return:
             A set of (LPG machine vertex, source machine vertex, partition_id)
        :rtype: set(tuple(LivePacketGatherMachineVertex,
            ~pacman.model.graphs.machine.MachineVertex, str))
        """
        return self.__targeted_lpgs

    @overrides(AbstractSplitterCommon.get_out_going_vertices)
    def get_out_going_vertices(self, partition_id):
        # There are none!
        return []

    @overrides(AbstractSplitterCommon.machine_vertices_for_recording)
    def machine_vertices_for_recording(self, variable_to_record):
        # Nothing to record here...
        return []

    @overrides(AbstractSplitterCommon.reset_called)
    def reset_called(self):
        self.__m_vertices_by_ethernet = dict()
        self.__targeted_lpgs = set()



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import struct
from spinn_utilities.log import FormatAdapter
from spinn_utilities.progress_bar import ProgressBar
from data_specification.constants import MAX_MEM_REGIONS
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.constants import BYTES_PER_WORD

logger = FormatAdapter(logging.getLogger(__name__))
_ONE_WORD = struct.Struct("<I")
REGION_HEADER_SIZE = 2 * BYTES_PER_WORD


def memory_map_on_host_chip_report():
    """
    Report on memory usage. Creates a report that states where in SDRAM
    each region is (read from machine).
    """
    directory_name = os.path.join(
        FecDataView.get_run_dir_path(), "memory_map_reports")
    if not os.path.exists(directory_name):
        os.makedirs(directory_name)

    transceiver = FecDataView.get_transceiver()
    dsg_targets = FecDataView.get_dsg_targets()
    progress = ProgressBar(
        dsg_targets.ds_n_cores(), "Writing memory map reports")
    for (x, y, p) in progress.over(dsg_targets.keys()):
        file_name = os.path.join(
            directory_name, f"memory_map_from_processor_{x}_{y}_{p}.txt")
        try:
            with open(file_name, "w", encoding="utf-8") as f:
                _describe_mem_map(f, transceiver, x, y, p)
        except IOError:
            logger.exception(
                "Generate_placement_reports: Can't open file {} for writing.",
                file_name)


def _describe_mem_map(f, txrx, x, y, p):
    """
    :param ~spinnman.transceiver.Transceiver txrx:
    """
    # Read the memory map data from the given core
    region_table_addr = _get_region_table_addr(txrx, x, y, p)
    memmap_data = txrx.read_memory(
        x, y, region_table_addr, BYTES_PER_WORD * MAX_MEM_REGIONS)

    # Convert the map to a human-readable description
    f.write("On chip data specification executor\n\n")
    for i in range(MAX_MEM_REGIONS):
        region_address, = _ONE_WORD.unpack_from(
            memmap_data, i * BYTES_PER_WORD)
        f.write(f"Region {i:d}:\n\t start address: 0x{region_address:x}\n\n")


def _get_region_table_addr(txrx, x, y, p):
    """
    :param ~spinnman.transceiver.Transceiver txrx:
    """
    user_0_addr = txrx.get_user_0_register_address_from_core(p)
    return txrx.read_word(x, y, user_0_addr) + REGION_HEADER_SIZE



# Copyright (c) 2014 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pacman.model.graphs.application import ApplicationEdge, ApplicationVertex
from pacman.model.graphs.application.abstract import (
    AbstractOneAppOneMachineVertex)
from .command_sender_machine_vertex import CommandSenderMachineVertex
from spinn_utilities.overrides import overrides


class CommandSender(
        AbstractOneAppOneMachineVertex):
    """
    A utility for sending commands to a vertex (possibly an external device)
    at fixed times in the simulation or in response to simulation events
    (e.g., starting and stopping).
    """

    def __init__(self, label):
        """
        :param str label: The label of this vertex
        """
        super().__init__(
            CommandSenderMachineVertex(label, self), label)

    def add_commands(
            self, start_resume_commands, pause_stop_commands,
            timed_commands, vertex_to_send_to):
        """
        Add commands to be sent down a given edge.

        :param iterable(MultiCastCommand) start_resume_commands:
            The commands to send when the simulation starts or resumes from
            pause
        :param iterable(MultiCastCommand) pause_stop_commands:
            The commands to send when the simulation stops or pauses after
            running
        :param iterable(MultiCastCommand) timed_commands:
            The commands to send at specific times
        :param ~pacman.model.graphs.AbstractVertex vertex_to_send_to:
            The vertex these commands are to be sent to
        """
        self._machine_vertex.add_commands(
            start_resume_commands, pause_stop_commands, timed_commands,
            vertex_to_send_to)

    def edges_and_partitions(self):
        """
        Construct application edges from this vertex to the app vertices
        that this vertex knows how to target (and has keys allocated for).

        :return: edges, partition IDs
        :rtype: tuple(list(~pacman.model.graphs.application.ApplicationEdge),
            list(str))
        """
        return self._machine_vertex.get_edges_and_partitions(
            self, ApplicationVertex, ApplicationEdge)

    @overrides(AbstractOneAppOneMachineVertex.get_fixed_key_and_mask)
    def get_fixed_key_and_mask(self, partition_id):
        return self._machine_vertex.get_fixed_key_and_mask(partition_id)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from spinnman.model import ExecutableTargets
from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from pacman.model.graphs import AbstractVirtual
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import (
    ExecutableNotFoundException)
from spinn_front_end_common.abstract_models import AbstractHasAssociatedBinary

logger = FormatAdapter(logging.getLogger(__name__))


def graph_binary_gatherer():
    """
    Extracts binaries to be executed.

    :rtype: ~spinnman.model.ExecutableTargets
    """
    gatherer = _GraphBinaryGatherer()
    # pylint: disable=protected-access
    return gatherer._run()


class _GraphBinaryGatherer(object):
    """
    Extracts binaries to be executed.
    """

    __slots__ = ["_exe_targets"]

    def __init__(self):
        self._exe_targets = ExecutableTargets()

    def _run(self):
        """
        :param ~pacman.model.placements.Placements placements:
        :rtype: ~spinnman.model.ExecutableTargets
        """
        progress = ProgressBar(
            FecDataView.get_n_placements(), "Finding binaries")
        for placement in progress.over(FecDataView.iterate_placemements()):
            self.__get_binary(placement)

        return self._exe_targets

    def __get_binary(self, placement):
        """
        :param ~pacman.model.placements.Placement placement:
        :param ~pacman.model.graphs.AbstractVertex vertex:
        """
        # if the vertex cannot be executed, ignore it
        vertex = placement.vertex
        if not isinstance(vertex, AbstractHasAssociatedBinary):
            if not isinstance(vertex, AbstractVirtual):
                logger.error(
                    "Vertex {} does not implement either "
                    "AbstractHasAssociatedBinary or AbstractVirtual. So it is "
                    "unclear if it should or should not have a binary", vertex)
            return

        # Get name of binary from vertex
        binary_name = vertex.get_binary_file_name()
        exec_type = vertex.get_binary_start_type()

        # Attempt to find this within search paths
        binary_path = FecDataView.get_executable_path(binary_name)
        if binary_path is None:
            raise ExecutableNotFoundException(binary_name)

        self._exe_targets.add_processor(
            binary_path, placement.x, placement.y, placement.p, exec_type)



# Copyright (c) 2015 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
from threading import Thread, Condition
from time import sleep
from spinn_utilities.log import FormatAdapter
from spinnman.messages.eieio.data_messages import (
    EIEIODataMessage, KeyPayloadDataElement)
from spinnman.messages.eieio import EIEIOType, AbstractEIEIOMessage
from spinnman.connections import ConnectionListener
from spinnman.connections.udp_packet_connections import (
    EIEIOConnection, UDPConnection)
from spinnman.messages.sdp.sdp_flag import SDPFlag
from spinnman.constants import SCP_SCAMP_PORT
from spinnman.messages.sdp.sdp_message import SDPMessage
from spinnman.messages.sdp.sdp_header import SDPHeader
from spinnman.utilities.utility_functions import reprogram_tag_to_listener
from spinnman.messages.eieio import (
    read_eieio_command_message, read_eieio_data_message)
from spinn_front_end_common.utilities.constants import NOTIFY_PORT
from spinn_front_end_common.utilities.database import DatabaseConnection
from spinn_front_end_common.utilities.exceptions import ConfigurationException

logger = FormatAdapter(logging.getLogger(__name__))

# The maximum number of 32-bit keys that will fit in a packet
_MAX_FULL_KEYS_PER_PACKET = 63

# The maximum number of 16-bit keys that will fit in a packet
_MAX_HALF_KEYS_PER_PACKET = 127

# The maximum number of 32-bit keys with payloads that will fit in a packet
_MAX_FULL_KEYS_PAYLOADS_PER_PACKET = 31

# Decoding of a single short value
_ONE_SHORT = struct.Struct("<H")

# The size of a RAW SCP OK response (includes 2 bytes of padding)
_SCP_OK_SIZE = 18

# The byte of the RAW SCP packet that contains the flags
_SCP_FLAGS_BYTE = 2

# The byte of the RAW SCP packet that contains the destination cpu
_SCP_DEST_CPU_BYTE = 4

# The expected flags from a RAW SCP packet in response
_SCP_RESPONSE_FLAGS = 7

# The expected destination cpu from a RAW SCP packet in repsonse
_SCP_RESPONSE_DEST = 0xFF


class LiveEventConnection(DatabaseConnection):
    """
    A connection for receiving and sending live events from and to SpiNNaker.

    .. note::
        This class is intended to be potentially usable from another
        process than the one that the simulator is present in.
    """
    __slots__ = [
        "_atom_id_to_key",
        "__error_keys",
        "__init_callbacks",
        "__key_to_atom_id_and_label",
        "__live_event_callbacks",
        "__live_packet_gather_label",
        "__pause_stop_callbacks",
        "__receive_labels",
        "__receiver_connection",
        "__receiver_listener",
        "__send_address_details",
        "__send_labels",
        "__sender_connection",
        "__start_resume_callbacks",
        "__simulator",
        "__spalloc_job",
        "__receiver_details",
        "__is_running",
        "__expect_scp_response",
        "__expect_scp_response_lock",
        "__scp_response_received"]

    def __init__(self, live_packet_gather_label, receive_labels=None,
                 send_labels=None, local_host=None, local_port=NOTIFY_PORT):
        """
        :param str live_packet_gather_label:
            The label of the vertex to which received events are being sent
        :param iterable(str) receive_labels:
            Labels of vertices from which live events will be received.
        :param iterable(str) send_labels:
            Labels of vertices to which live events will be sent
        :param str local_host:
            Optional specification of the local hostname or IP address of the
            interface to listen on
        :param int local_port:
            Optional specification of the local port to listen on. Must match
            the port that the toolchain will send the notification on (19999
            by default)
        """
        # pylint: disable=too-many-arguments
        super().__init__(
            self.__do_start_resume, self.__do_stop_pause,
            local_host=local_host, local_port=local_port)

        self.add_database_callback(self.__read_database_callback)

        self.__live_packet_gather_label = live_packet_gather_label
        self.__receive_labels = (
            list(receive_labels) if receive_labels is not None else None)
        self.__send_labels = (
            list(send_labels) if send_labels is not None else None)
        self.__sender_connection = None
        self.__send_address_details = dict()
        # Also used by SpynnakerPoissonControlConnection
        self._atom_id_to_key = dict()
        self.__key_to_atom_id_and_label = dict()
        self.__live_event_callbacks = list()
        self.__start_resume_callbacks = dict()
        self.__pause_stop_callbacks = dict()
        self.__init_callbacks = dict()
        self.__receiver_details = list()
        if receive_labels is not None:
            for label in receive_labels:
                self.__live_event_callbacks.append(list())
                self.__start_resume_callbacks[label] = list()
                self.__pause_stop_callbacks[label] = list()
                self.__init_callbacks[label] = list()
        if send_labels is not None:
            for label in send_labels:
                self.__start_resume_callbacks[label] = list()
                self.__pause_stop_callbacks[label] = list()
                self.__init_callbacks[label] = list()
        self.__receiver_listener = None
        self.__receiver_connection = None
        self.__error_keys = set()
        self.__is_running = False
        self.__expect_scp_response = False
        self.__expect_scp_response_lock = Condition()
        self.__scp_response_received = None

    def add_send_label(self, label):
        if self.__send_labels is None:
            self.__send_labels = list()
        if label not in self.__send_labels:
            self.__send_labels.append(label)
        if label not in self.__start_resume_callbacks:
            self.__start_resume_callbacks[label] = list()
            self.__pause_stop_callbacks[label] = list()
            self.__init_callbacks[label] = list()

    def add_receive_label(self, label):
        if self.__receive_labels is None:
            self.__receive_labels = list()
        if label not in self.__receive_labels:
            self.__receive_labels.append(label)
            self.__live_event_callbacks.append(list())
        if label not in self.__start_resume_callbacks:
            self.__start_resume_callbacks[label] = list()
            self.__pause_stop_callbacks[label] = list()
            self.__init_callbacks[label] = list()

    def add_init_callback(self, label, init_callback):
        """
        Add a callback to be called to initialise a vertex.

        :param str label:
            The label of the vertex to be notified about. Must be one of the
            vertices listed in the constructor
        :param init_callback: A function to be called to initialise the
            vertex. This should take as parameters the label of the vertex,
            the number of neurons in the population, the run time of the
            simulation in milliseconds, and the simulation timestep in
            milliseconds
        :type init_callback: callable(str, int, float, float) -> None
        """
        self.__init_callbacks[label].append(init_callback)

    def add_receive_callback(self, label, live_event_callback,
                             translate_key=True):
        """
        Add a callback for the reception of live events from a vertex.

        :param str label: The label of the vertex to be notified about.
            Must be one of the vertices listed in the constructor
        :param live_event_callback: A function to be called when events are
            received. This should take as parameters the label of the vertex,
            the simulation timestep when the event occurred, and an
            array-like of atom IDs.
        :type live_event_callback: callable(str, int, list(int)) -> None
        :param bool translate_key:
            True if the key is to be converted to an atom ID, False if the
            key should stay a key
        """
        label_id = self.__receive_labels.index(label)
        logger.info("Receive callback {} registered to label {}",
                    live_event_callback, label)
        self.__live_event_callbacks[label_id].append(
            (live_event_callback, translate_key))

    def add_start_callback(self, label, start_callback):
        """
        Add a callback for the start of the simulation.

        :param start_callback: A function to be called when the start
            message has been received. This function should take the label of
            the referenced vertex, and an instance of this class, which can
            be used to send events
        :type start_callback: callable(str, LiveEventConnection) -> None
        :param str label: the label of the function to be sent
        """
        logger.warning(
            "the method 'add_start_callback(label, start_callback)' is in "
            "deprecation, and will be replaced with the method "
            "'add_start_resume_callback(label, start_resume_callback)' in a "
            "future release.")
        self.add_start_resume_callback(label, start_callback)

    def add_start_resume_callback(self, label, start_resume_callback):
        """
        Add a callback for the start and resume state of the simulation.

        :param str label: the label of the function to be sent
        :param start_resume_callback: A function to be called when the start
            or resume message has been received. This function should take
            the label of the referenced vertex, and an instance of this
            class, which can be used to send events.
        :type start_resume_callback: callable(str, LiveEventConnection) -> None
        """
        self.__start_resume_callbacks[label].append(start_resume_callback)

    def add_pause_stop_callback(self, label, pause_stop_callback):
        """
        Add a callback for the pause and stop state of the simulation.

        :param str label: the label of the function to be sent
        :param pause_stop_callback: A function to be called when the pause
            or stop message has been received. This function should take the
            label of the referenced  vertex, and an instance of this class,
            which can be used to send events.
        :type pause_stop_callback: callable(str, LiveEventConnection) -> None
        """
        self.__pause_stop_callbacks[label].append(pause_stop_callback)

    def __read_database_callback(self, db_reader):
        """
        :param DatabaseReader db_reader:
        """
        self.__handle_possible_rerun_state()

        vertex_sizes = dict()
        run_time_ms = db_reader.get_configuration_parameter_value(
            "runtime")
        machine_timestep_ms = db_reader.get_configuration_parameter_value(
            "machine_time_step") / 1000.0

        if self.__send_labels is not None:
            self.__init_sender(db_reader, vertex_sizes)

        if self.__receive_labels is not None:
            self.__init_receivers(db_reader, vertex_sizes)

        for label, vertex_size in vertex_sizes.items():
            for init_callback in self.__init_callbacks[label]:
                init_callback(
                    label, vertex_size, run_time_ms, machine_timestep_ms)

    def __init_sender(self, db, vertex_sizes):
        """
        :param DatabaseReader db:
        :param dict(str,int) vertex_sizes:
        """
        if self.__sender_connection is None:
            job = db.get_job()
            if job:
                self.__sender_connection = job.open_listener_connection()
            else:
                self.__sender_connection = EIEIOConnection()
        for label in self.__send_labels:
            self.__send_address_details[label] = self.__get_live_input_details(
                db, label)
            self._atom_id_to_key[label] = db.get_atom_id_to_key_mapping(
                label)
            vertex_sizes[label] = len(self._atom_id_to_key[label])

    def __init_receivers(self, db, vertex_sizes):
        """
        :param DatabaseReader db:
        :param dict(str,int) vertex_sizes:
        """
        # Set up a single connection for receive
        if self.__receiver_connection is None:
            job = db.get_job()
            if job:
                self.__receiver_connection = job.open_listener_connection()
            else:
                self.__receiver_connection = UDPConnection()
        receivers = set()
        for label_id, label in enumerate(self.__receive_labels):
            _, port, board_address, tag, x, y = self.__get_live_output_details(
                db, label)

            # Update the tag if not already done
            if (board_address, port, tag) not in receivers:
                receivers.add((board_address, port, tag))
                self.__receiver_details.append((x, y, tag, board_address))

            logger.info(
                "Listening for traffic from {} on board {} on {}:{}",
                label, board_address,
                self.__receiver_connection.local_ip_address,
                self.__receiver_connection.local_port)

            key_to_atom_id = db.get_key_to_atom_id_mapping(label)
            for key, atom_id in key_to_atom_id.items():
                self.__key_to_atom_id_and_label[key] = (atom_id, label_id)
            vertex_sizes[label] = len(key_to_atom_id)

        # Last of all, set up the listener for packets
        # NOTE: Has to be done last as otherwise will receive SCP messages
        # sent above!
        if self.__receiver_listener is None:
            self.__receiver_listener = ConnectionListener(
                self.__receiver_connection)
            self.__receiver_listener.add_callback(self.__do_receive_packet)
            self.__receiver_listener.start()
            self.__send_tag_messages_now()

    def __get_live_input_details(self, db_reader, send_label):
        """
        :param DatabaseReader db_reader:
        :param str send_label:
        :rtype: tuple(int,int,int,str or None)
        """
        x, y, p = db_reader.get_placements(send_label)[0]

        ip_address = db_reader.get_ip_address(x, y)
        return x, y, p, ip_address

    def __get_live_output_details(self, db_reader, receive_label):
        host, port, strip_sdp, board_address, tag, chip_x, chip_y = \
            db_reader.get_live_output_details(
                receive_label, self.__live_packet_gather_label)
        if host is None:
            raise ConfigurationException(
                f"no live output tag found for {receive_label} in app graph")
        if not strip_sdp:
            raise receive_label("Currently, only IP tags which strip the SDP "
                                "headers are supported")
        return host, port, board_address, tag, chip_x, chip_y

    def __handle_possible_rerun_state(self):
        # reset from possible previous calls
        if self.__sender_connection is not None:
            self.__sender_connection.close()
            self.__sender_connection = None
        if self.__receiver_listener is not None:
            self.__receiver_listener.close()
            self.__receiver_listener = None
        if self.__receiver_connection is not None:
            self.__receiver_connection.close()
            self.__receiver_connection = None

    def __launch_thread(self, kind, label, callback):
        thread = Thread(
            target=callback, args=(label, self),
            name=(f"{kind} callback thread for live_event_connection "
                  f"{self._local_port}:{self._local_ip_address}"))
        thread.start()

    def __do_start_resume(self):
        for label, callbacks in self.__start_resume_callbacks.items():
            for callback in callbacks:
                self.__launch_thread("start_resume", label, callback)
        if not self.__is_running:
            self.__is_running = True
            thread = Thread(target=self.__send_tag_messages_thread)
            thread.start()

    def __do_stop_pause(self):
        self.__is_running = False
        for label, callbacks in self.__pause_stop_callbacks.items():
            for callback in callbacks:
                self.__launch_thread("pause_stop", label, callback)

    def __send_tag_messages_thread(self):
        if self.__receiver_connection is None:
            return
        while self.__is_running:
            sleep(10.0)
            if self.__is_running:
                self.__send_tag_messages_now()

    def __send_tag_messages_now(self):
        indirect = hasattr(self.__receiver_connection, "update_tag")
        for (x, y, tag, board_address) in self.__receiver_details:
            with self.__expect_scp_response_lock:
                self.__scp_response_received = None
                self.__expect_scp_response = True
                if indirect:
                    self.__receiver_connection.update_tag(
                        x, y, tag, do_receive=False)
                    # No port trigger necessary; proxied already
                else:
                    reprogram_tag_to_listener(
                        self.__receiver_connection, x, y, board_address,
                        tag, read_response=False)
                while self.__scp_response_received is None:
                    self.__expect_scp_response_lock.wait(timeout=1.0)

    def __handle_scp_packet(self, data):
        with self.__expect_scp_response_lock:

            # SCP unexpected
            if not self.__expect_scp_response:
                return False

            if (len(data) == _SCP_OK_SIZE and
                    data[_SCP_FLAGS_BYTE] == _SCP_RESPONSE_FLAGS and
                    data[_SCP_DEST_CPU_BYTE] == _SCP_RESPONSE_DEST):
                self.__scp_response_received = data
                self.__expect_scp_response = False
                return True
            return False

    def __do_receive_packet(self, data):
        if self.__handle_scp_packet(data):
            return

        logger.debug("Received packet")
        try:
            header = _ONE_SHORT.unpack_from(data)[0]
            if header & 0xC000 == 0x4000:
                return read_eieio_command_message(data, 0)
            packet = read_eieio_data_message(data, 0)
            if packet.eieio_header.is_time:
                self.__handle_time_packet(packet)
            else:
                self.__handle_no_time_packet(packet)

        # pylint: disable=broad-except
        except Exception:
            logger.warning("problem handling received packet", exc_info=True)

    def __handle_time_packet(self, packet):
        key_times_labels = dict()
        atoms_times_labels = dict()
        while packet.is_next_element:
            element = packet.next_element
            time = element.payload
            key = element.key
            if key in self.__key_to_atom_id_and_label:
                atom_id, label_id = self.__key_to_atom_id_and_label[key]
                if time not in key_times_labels:
                    key_times_labels[time] = dict()
                    atoms_times_labels[time] = dict()
                if label_id not in key_times_labels[time]:
                    key_times_labels[time][label_id] = list()
                    atoms_times_labels[time][label_id] = list()
                key_times_labels[time][label_id].append(key)
                atoms_times_labels[time][label_id].append(atom_id)
            else:
                self.__handle_unknown_key(key)

        for time in key_times_labels:
            for label_id in key_times_labels[time]:
                label = self.__receive_labels[label_id]
                for c_back, use_atom in self.__live_event_callbacks[label_id]:
                    if use_atom:
                        c_back(label, time, atoms_times_labels[time][label_id])
                    else:
                        c_back(label, time, key_times_labels[time][label_id])

    def __handle_no_time_packet(self, packet):
        while packet.is_next_element:
            element = packet.next_element
            key = element.key
            if key in self.__key_to_atom_id_and_label:
                atom_id, label_id = self.__key_to_atom_id_and_label[key]
                label = self.__receive_labels[label_id]
                for c_back, use_atom in self.__live_event_callbacks[label_id]:
                    if isinstance(element, KeyPayloadDataElement):
                        if use_atom:
                            c_back(label, atom_id, element.payload)
                        else:
                            c_back(label, key, element.payload)
                    else:
                        if use_atom:
                            c_back(label, atom_id)
                        else:
                            c_back(label, key)
            else:
                self.__handle_unknown_key(key)

    def __handle_unknown_key(self, key):
        if key not in self.__error_keys:
            self.__error_keys.add(key)
            logger.warning("Received unexpected key {}", key)

    def send_event(self, label, atom_id, send_full_keys=False):
        """
        Send an event from a single atom.

        :param str label:
            The label of the vertex from which the event will originate
        :param int atom_id: The ID of the atom sending the event
        :param bool send_full_keys:
            Determines whether to send full 32-bit keys, getting the key for
            each atom from the database, or whether to send 16-bit atom IDs
            directly
        """
        self.send_events(label, [atom_id], send_full_keys)

    def send_events(self, label, atom_ids, send_full_keys=False):
        """
        Send a number of events.

        :param str label:
            The label of the vertex from which the events will originate
        :param list(int) atom_ids: array-like of atom IDs sending events
        :param bool send_full_keys:
            Determines whether to send full 32-bit keys, getting the key for
            each atom from the database, or whether to send 16-bit atom IDs
            directly
        """
        max_keys = _MAX_HALF_KEYS_PER_PACKET
        msg_type = EIEIOType.KEY_16_BIT
        if send_full_keys:
            max_keys = _MAX_FULL_KEYS_PER_PACKET
            msg_type = EIEIOType.KEY_32_BIT

        pos = 0
        x, y, p, ip_address = self.__send_address_details[label]
        while pos < len(atom_ids):
            message = EIEIODataMessage.create(msg_type)
            events_in_packet = 0
            while pos < len(atom_ids) and events_in_packet < max_keys:
                key = atom_ids[pos]
                if send_full_keys:
                    key = self._atom_id_to_key[label][key]
                message.add_key(key)
                pos += 1
                events_in_packet += 1

            self._send(message, x, y, p, ip_address)

    def send_event_with_payload(self, label, atom_id, payload):
        """
        Send an event with a payload from a single atom.

        :param str label:
            The label of the vertex from which the event will originate
        :param int atom_id: The ID of the atom sending the event
        :param int payload: The payload to send
        """
        self.send_events_with_payloads(label, [(atom_id, payload)])

    def send_events_with_payloads(self, label, atom_ids_and_payloads):
        """
        Send a number of events with payloads.

        :param str label:
            The label of the vertex from which the events will originate
        :param list(tuple(int,int)) atom_ids_and_payloads:
            array-like of tuples of atom IDs sending events with their payloads
        """
        msg_type = EIEIOType.KEY_PAYLOAD_32_BIT
        max_keys = _MAX_FULL_KEYS_PAYLOADS_PER_PACKET
        pos = 0
        x, y, p, ip_address = self.__send_address_details[label]
        while pos < len(atom_ids_and_payloads):
            message = EIEIODataMessage.create(msg_type)
            events = 0
            while pos < len(atom_ids_and_payloads) and events < max_keys:
                key, payload = atom_ids_and_payloads[pos]
                key = self._atom_id_to_key[label][key]
                message.add_key_and_payload(key, payload)
                pos += 1
                events += 1

            self._send(message, x, y, p, ip_address)

    def send_eieio_message(self, message, label):
        """
        Send an EIEIO message (using one-way the live input) to the
        vertex with the given label.

        :param ~spinnman.messages.eieio.AbstractEIEIOMessage message:
            The EIEIO message to send
        :param str label: The label of the receiver machine vertex
        """
        target = self.__send_address_details[label]
        if target is None:
            return
        x, y, p, ip_address = target
        self._send(message, x, y, p, ip_address)

    def _send(self, message: AbstractEIEIOMessage, x, y, p, ip_address):
        """
        Send an EIEIO message to a particular core.

        :param ~spinnman.messages.eieio.AbstractEIEIOMessage message:
            The EIEIO message to send
        :param int x: Destination chip X coordinate
        :param int y: Destination chip Y coordinate
        :param int p: Destination core number
        :param str ip_address:
            What Ethernet-enabled chip to send via (or rather its IP address)
        """
        # Create an SDP message - no reply so source is unimportant
        # SDP port can be anything except 0 as the target doesn't care
        sdp_message = SDPMessage(
            SDPHeader(
                flags=SDPFlag.REPLY_NOT_EXPECTED, tag=0,
                destination_port=1, destination_cpu=p,
                destination_chip_x=x, destination_chip_y=y,
                source_port=0, source_cpu=0,
                source_chip_x=0, source_chip_y=0),
            data=message.bytestring)
        self.__sender_connection.send_to(
            # Prefix padding: two empty bytes
            b'\0\0' + sdp_message.bytestring,
            (ip_address, SCP_SCAMP_PORT))

    def close(self):
        self.__is_running = False
        self.__handle_possible_rerun_state()
        super().close()



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import math
import os
import struct
from collections import defaultdict
from spinn_utilities.config_holder import get_config_bool
from spinn_utilities.find_max_success import find_max_success
from spinn_utilities.progress_bar import ProgressBar
from spinn_machine import Machine, MulticastRoutingEntry
from pacman.exceptions import (
    PacmanAlgorithmFailedToGenerateOutputsException,
    PacmanElementAllocationException, MinimisationFailedError)
from pacman.model.routing_tables import (
    MulticastRoutingTables, UnCompressedMulticastRoutingTable,
    CompressedMulticastRoutingTable)
from pacman.utilities.algorithm_utilities.routes_format import format_route
from spinn_front_end_common.abstract_models import (
    AbstractSupportsBitFieldRoutingCompression)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.helpful_functions import n_word_struct
from spinn_front_end_common.utilities.constants import (
    BYTES_PER_WORD, BYTES_PER_4_WORDS)
from spinn_front_end_common.utilities.report_functions.\
    bit_field_compressor_report import (
        generate_provenance_item)
from pacman.operations.router_compressors.pair_compressor import (
    _PairCompressor)

_REPORT_FOLDER_NAME = "router_compressor_with_bitfield"


def host_based_bit_field_router_compressor():
    """
    Entry point when using the PACMANAlgorithmExecutor.

    :return: compressed routing table entries
    :rtype: ~pacman.model.routing_tables.MulticastRoutingTables
    """
    routing_tables = FecDataView.get_uncompressed().routing_tables
    # create progress bar
    progress = ProgressBar(
        len(routing_tables) * 2,
        "Compressing routing Tables with bitfields in host")

    # create report
    if get_config_bool(
            "Reports", "write_router_compression_with_bitfield_report"):
        report_folder_path = generate_report_path()
    else:
        report_folder_path = None

    # compressed router table
    compressed_pacman_router_tables = MulticastRoutingTables()

    key_atom_map = generate_key_to_atom_map()

    most_costly_cores = defaultdict(lambda: defaultdict(int))
    for partition in FecDataView.iterate_partitions():
        for edge in partition.edges:
            splitter = edge.post_vertex.splitter
            for vertex, _ in splitter.get_source_specific_in_coming_vertices(
                    partition.pre_vertex, partition.identifier):
                place = FecDataView.get_placement_of_vertex(vertex)
                most_costly_cores[place.xy][place.p] += 1

    # start the routing table choice conversion
    for router_table in progress.over(routing_tables):
        start_compression_selection_process(
            router_table, report_folder_path,
            most_costly_cores, compressed_pacman_router_tables, key_atom_map)

    # return compressed tables
    return compressed_pacman_router_tables


def generate_report_path():
    """
    :rtype: str
    """
    report_folder_path = os.path.join(
        FecDataView.get_run_dir_path(), _REPORT_FOLDER_NAME)
    if not os.path.exists(report_folder_path):
        os.mkdir(report_folder_path)
    return report_folder_path


def generate_key_to_atom_map():
    """
    *THIS IS NEEDED* due to the link from key to edge being lost.

    :return: key to atom map based of key to n atoms
    :rtype: dict(int,int)
    """
    # build key to n atoms map
    routing_infos = FecDataView.get_routing_infos()
    key_to_n_atoms_map = dict()
    for partition in FecDataView.iterate_partitions():
        for vertex in partition.pre_vertex.splitter.get_out_going_vertices(
                partition.identifier):
            key = routing_infos.get_first_key_from_pre_vertex(
                vertex, partition.identifier)
            key_to_n_atoms_map[key] = vertex.vertex_slice.n_atoms
    return key_to_n_atoms_map


def start_compression_selection_process(
        router_table, report_folder_path,
        most_costly_cores, compressed_pacman_router_tables, key_atom_map):
    """
    Entrance method for doing on host compression. Can be used as a
    public method for other compressors.

    :param router_table: the routing table in question to compress
    :type router_table:
        ~pacman.model.routing_tables.UnCompressedMulticastRoutingTable
    :param report_folder_path: the report folder base address
    :type report_folder_path: str or None
    :param dict(dict(int)) most_costly_cores:
        Map of chip x, y to processors to count of incoming on processor
    :param dict(int,int) key_atom_map: key to atoms map
        should be allowed to handle per time step
    """
    compressor = _HostBasedBitFieldRouterCompressor()
    # pylint: disable=protected-access
    compressor._run(
        router_table, report_folder_path, most_costly_cores,
        compressed_pacman_router_tables, key_atom_map)


class _BitFieldData(object):
    __slots__ = [
        # bit_field data
        "bit_field",
        # Key this applies to
        "master_pop_key",
        # address of n_atoms word to wrote back merged flags
        "n_atoms_address",
        # Word that holds merged: 1; all_ones: 1; n_atoms: 30;
        "n_atoms_word",
        # P cooridnate of processor this applies to
        "processor_id",
        # index of this entry in the "sorted_list"
        "sort_index",
        # The shift to get the core id
        "core_shift",
        # The number of atoms per core
        "n_atoms_per_core"
    ]

    def __init__(self, processor_id, bit_field, master_pop_key,
                 n_atoms_address, n_atoms_word, core_shift, n_atoms_per_core):
        self.processor_id = processor_id
        self.bit_field = bit_field
        self.master_pop_key = master_pop_key
        self.n_atoms_address = n_atoms_address
        self.n_atoms_word = n_atoms_word
        self.core_shift = core_shift
        self.n_atoms_per_core = n_atoms_per_core
        self.sort_index = None

    def __str__(self):
        return f"{self.processor_id} {self.master_pop_key} {self.bit_field}"

    def bit_field_as_bit_array(self):
        """
        Convert the bitfield into an array of bits like bit_field.c
        """
        return [(word >> i) & 1
                for word in self.bit_field
                for i in range(32)]


class _HostBasedBitFieldRouterCompressor(object):
    """
    Host-based fancy router compressor using the bitfield filters of the
    cores. Compresses bitfields and router table entries together as
    much as feasible.
    """

    __slots__ = [
        # List of the entries from the highest successful midpoint
        "_best_routing_entries",
        # The highest successful midpoint
        "_best_midpoint",
        # Dict of lists of the original bitfields by key
        "_bit_fields_by_key",
        # Record of which midpoints where tired and with what result
        "_compression_attempts",
        # Number of bitfields for this core
        "_n_bitfields"
    ]

    # max entries that can be used by the application code
    _MAX_SUPPORTED_LENGTH = 1023

    # the amount of time each attempt at router compression can be allowed to
    #  take (in seconds)
    # _DEFAULT_TIME_PER_ITERATION = 5 * 60
    _DEFAULT_TIME_PER_ITERATION = 10

    # report name
    _REPORT_NAME = "router_{}_{}.rpt"

    # key id for the initial entry
    _ORIGINAL_ENTRY = 0

    # key id for the bitfield entries
    _ENTRIES = 1

    # size of a filter info in bytes (key, n words, pointer)
    _SIZE_OF_FILTER_INFO_IN_BYTES = 12

    # bit to mask a bit
    _BIT_MASK = 1

    # mask for neuron level
    _NEURON_LEVEL_MASK = 0xFFFFFFFF

    # to remove the first   merged: 1; and all_ones: 1;
    N_ATOMS_MASK = 0x3FFFFFFF
    MERGED_SETTER = 0x80000000

    # struct for performance requirements.
    _FOUR_WORDS = struct.Struct("<IIII")

    # for router report
    _LOWER_16_BITS = 0xFFFF

    # rob paul to pay sam threshold starting point at 1ms time step
    _N_PACKETS_PER_SECOND = 100000

    # convert between milliseconds and second
    _MS_TO_SEC = 1000

    # fraction effect for threshold
    _THRESHOLD_FRACTION_EFFECT = 2

    # Number of bits per word as an int
    _BITS_PER_WORD = 32

    def __init__(self):
        self._best_routing_entries = None
        self._best_midpoint = -1
        self._bit_fields_by_key = None
        self._compression_attempts = dict()
        self._n_bitfields = None

    def get_bit_field_sdram_base_addresses(self, chip_x, chip_y):
        """
        :param int chip_x:
        :param int chip_y:
        """
        # pylint: disable=too-many-arguments, unused-argument
        # locate the bitfields in a chip level scope
        base_addresses = dict()
        for placement in FecDataView.iterate_placements_by_xy_and_type(
                chip_x, chip_y, AbstractSupportsBitFieldRoutingCompression):
            vertex = placement.vertex
            base_addresses[placement.p] = vertex.bit_field_base_address(
                FecDataView.get_placement_of_vertex(vertex))
        return base_addresses

    def _run(
            self, router_table, report_folder_path, most_costly_cores,
            compressed_pacman_router_tables, key_atom_map):
        """
        Entrance method for doing on host compression. Can be used as a
        public method for other compressors.

        :param router_table: the routing table in question to compress
        :type router_table:
            ~pacman.model.routing_tables.UnCompressedMulticastRoutingTable
        :param report_folder_path: the report folder base address
        :type report_folder_path: str or None
        :param dict(dict(int)) most_costly_cores:
            Map of chip x, y to processors to count of incoming on processor
        :param compressed_pacman_router_tables:
            a data holder for compressed tables
        :type compressed_pacman_router_tables:
            ~pacman.model.routing_tables.MulticastRoutingTables
        :param dict(int,int) key_atom_map: key to atoms map
            should be allowed to handle per time step
        """
        # Find the processors that have bitfield data and where it is
        bit_field_chip_base_addresses = (
            self.get_bit_field_sdram_base_addresses(
                router_table.x, router_table.y))

        # read in bitfields.
        self._read_in_bit_fields(
            router_table.x, router_table.y, bit_field_chip_base_addresses,
            most_costly_cores)

        # execute binary search
        self._start_binary_search(router_table, key_atom_map)

        # add final to compressed tables:
        # self._best_routing_table is a list of entries
        best_router_table = CompressedMulticastRoutingTable(
            router_table.x, router_table.y)

        for entry in self._best_routing_entries:
            best_router_table.add_multicast_routing_entry(
                entry.to_MulticastRoutingEntry())

        compressed_pacman_router_tables.add_routing_table(best_router_table)

        # remove bitfields from cores that have been merged into the
        # router table
        self._remove_merged_bitfields_from_cores(
            router_table.x, router_table.y)

        # create report file if required
        if report_folder_path:
            report_file_path = os.path.join(
                report_folder_path,
                self._REPORT_NAME.format(router_table.x, router_table.y))
            with open(report_file_path, "w", encoding="utf-8") as report_out:
                self._create_table_report(router_table, report_out)

        generate_provenance_item(
            router_table.x, router_table.y, self._best_midpoint)

    def _convert_bitfields_into_router_table(
            self, router_table, mid_point, key_to_n_atoms_map):
        """
        Converts the bitfield into router table entries for compression,
        based off the entry located in the original router table.

        :param router_table: the original routing table
        :type router_table:
            ~pacman.model.routing_tables.UnCompressedMulticastRoutingTable
        :param int mid_point: cut-off for bitfields to use
        :param dict(int,int) key_to_n_atoms_map:
        :return: routing tables.
        :rtype: ~pacman.model.routing_tables.AbstractMulticastRoutingTable
        """
        new_table = UnCompressedMulticastRoutingTable(
            router_table.x, router_table.y)

        # go through the routing tables and convert when needed
        for original_entry in router_table.multicast_routing_entries:
            base_key = original_entry.routing_entry_key
            if base_key not in self._bit_fields_by_key:
                continue
            n_neurons = key_to_n_atoms_map[base_key]
            core_map = dict()
            entry_links = original_entry.link_ids
            # Assume all neurons for each processor to be kept
            for processor_id in original_entry.processor_ids:
                core_map[processor_id] = [1] * n_neurons
            # For those below the midpoint use the bitfield data
            for bf_data in self._bit_fields_by_key[base_key]:
                if bf_data.sort_index < mid_point:
                    core_map[bf_data.processor_id] = (
                        bf_data.bit_field_as_bit_array())
            # Add an Entry for each neuron
            for neuron in range(0, n_neurons):
                processors = list()
                for processor_id in original_entry.processor_ids:
                    if core_map[processor_id][neuron]:
                        processors.append(processor_id)
                # build new entry for this neuron and add to table
                new_table.add_multicast_routing_entry(MulticastRoutingEntry(
                    routing_entry_key=base_key + neuron,
                    mask=self._NEURON_LEVEL_MASK, link_ids=entry_links,
                    defaultable=False, processor_ids=processors))

        # return the bitfield tables and the reduced original table
        return new_table

    def _bit_for_neuron_id(self, bit_field, neuron_id):
        """
        Locate the bit for the neuron in the bitfield.

        :param list(int) bit_field:
            the block of words which represent the bitfield
        :param int neuron_id: the neuron id to find the bit in the bitfield
        :return: the bit
        """
        word_id = int(neuron_id // self._BITS_PER_WORD)
        bit_in_word = neuron_id % self._BITS_PER_WORD
        flag = (bit_field[word_id] >> bit_in_word) & self._BIT_MASK
        return flag

    def _read_in_bit_fields(
            self, chip_x, chip_y, bit_field_chip_base_addresses,
            most_costly_cores):
        """
        Read in the bitfields from the cores.

        :param int chip_x: chip x coordinate
        :param int chip_y: chip y coordinate
        :param dict(dict(int)) most_costly_cores:
            Map of chip x, y to processors to count of incoming on processor
        :param dict(int,int) bit_field_chip_base_addresses:
            maps core id to base address
        :return: dict of lists of processor id to bitfields.
        :rtype: tuple(dict(int,list(_BitFieldData)), list(_BitFieldData))
        """

        # data holder
        self._bit_fields_by_key = defaultdict(list)
        bit_fields_by_coverage = defaultdict(list)
        processor_coverage_by_bitfield = defaultdict(list)

        # read in for each app vertex that would have a bitfield
        for processor_id in bit_field_chip_base_addresses.keys():
            bit_field_base_address = (
                bit_field_chip_base_addresses[processor_id])

            # from filter_region_t read how many bitfields there are
            # n_filters then array of filters
            n_filters = FecDataView.get_transceiver().read_word(
                chip_x, chip_y, bit_field_base_address, BYTES_PER_WORD)
            reading_address = bit_field_base_address + BYTES_PER_WORD

            # read in each bitfield
            for _ in range(0, n_filters):
                master_pop_key, n_atoms_word, n_per_core_word, read_pointer = \
                    self._FOUR_WORDS.unpack(FecDataView.read_memory(
                        chip_x, chip_y, reading_address, BYTES_PER_4_WORDS))
                n_atoms_address = reading_address + BYTES_PER_WORD
                reading_address += BYTES_PER_4_WORDS

                # merged: 1; all_ones: 1; n_atoms: 30;
                atoms = n_atoms_word & self.N_ATOMS_MASK
                # get bitfield words
                n_words_to_read = math.ceil(atoms / self._BITS_PER_WORD)

                bit_field = n_word_struct(n_words_to_read).unpack(
                    FecDataView.read_memory(
                        chip_x, chip_y, read_pointer,
                        n_words_to_read * BYTES_PER_WORD))

                # sorted by best coverage of redundant packets
                data = _BitFieldData(
                    processor_id, bit_field, master_pop_key, n_atoms_address,
                    n_atoms_word, n_per_core_word & 0x1F, n_per_core_word >> 5)
                as_array = data.bit_field_as_bit_array()
                # Number of fields in the array that are zero instead of one
                n_redundant_packets = len(as_array) - sum(as_array)

                bit_fields_by_coverage[n_redundant_packets].append(data)
                processor_coverage_by_bitfield[processor_id].append(
                    n_redundant_packets)

                # add to the bitfields tracker
                self._bit_fields_by_key[master_pop_key].append(data)

        # use the ordered process to find the best ones to do first
        self._order_bit_fields(
            bit_fields_by_coverage, most_costly_cores, chip_x, chip_y,
            processor_coverage_by_bitfield)

    def _order_bit_fields(
            self, bit_fields_by_coverage, most_costly_cores, chip_x, chip_y,
            processor_coverage_by_bitfield):
        """
        Order the bit fields by redundancy setting the sorted index.

        Also counts the bitfields setting _n_bitfields.

        :param dict(int,list(_BitFieldData)) bit_fields_by_coverage:
        :param dict(dict(int)) most_costly_cores:
            Map of chip x, y to processors to count of incoming on processor
        :param int chip_x:
        :param int chip_y:
        :param dict(int,list(int)) processor_coverage_by_bitfield:
        """
        sort_index = 0

        # get cores that are the most likely to have the worst time and order
        #  bitfields accordingly
        cores = list(most_costly_cores[chip_x, chip_y].keys())
        cores = sorted(
            cores, key=lambda k: most_costly_cores[chip_x, chip_y][k],
            reverse=True)

        # only add bit fields from the worst affected cores, which will
        # build as more and more are taken to make the worst a collection
        # instead of a individual
        cores_to_add_for = list()
        for worst_core_id in range(0, len(cores) - 1):

            # determine how many of the worst to add before it balances with
            # next one
            cores_to_add_for.append(cores[worst_core_id])
            diff = (
                most_costly_cores[chip_x, chip_y][cores[worst_core_id]] -
                most_costly_cores[chip_x, chip_y][cores[worst_core_id + 1]])

            # order over most effective bitfields
            coverage = processor_coverage_by_bitfield[cores[worst_core_id]]
            coverage.sort(reverse=True)

            # cycle till at least the diff is covered
            covered = 0
            for redundant_packet_count in coverage:
                to_delete = list()
                for bit_field_data in bit_fields_by_coverage[
                        redundant_packet_count]:
                    if bit_field_data.processor_id in cores_to_add_for:
                        if covered < diff:
                            bit_field_data.sort_index = sort_index
                            sort_index += 1
                            to_delete.append(bit_field_data)
                            covered += 1
                for bit_field_data in to_delete:
                    bit_fields_by_coverage[redundant_packet_count].remove(
                        bit_field_data)

        # take left overs
        coverage_levels = list(bit_fields_by_coverage.keys())
        coverage_levels.sort(reverse=True)
        for coverage_level in coverage_levels:
            for bit_field_data in bit_fields_by_coverage[coverage_level]:
                bit_field_data.sort_index = sort_index
                sort_index += 1

        self._n_bitfields = sort_index

    def _start_binary_search(
            self, router_table, key_atom_map):
        """
        Start binary search of the merging of bitfield to router table.

        :param ~.UnCompressedMulticastRoutingTable router_table:
            uncompressed router table
        :param dict(int,int) key_atom_map: map from key to atoms
        """
        # try first just uncompressed. so see if its possible
        try:
            self._best_routing_entries = self._run_algorithm(router_table)
            self._best_midpoint = 0
            self._compression_attempts[0] = "succcess"
        except MinimisationFailedError as e:
            raise PacmanAlgorithmFailedToGenerateOutputsException(
                "host bitfield router compressor can't compress the "
                "uncompressed routing tables, regardless of bitfield merging. "
                "System is fundamentally flawed here") from e

        find_max_success(self._n_bitfields, functools.partial(
            self._binary_search_check, routing_table=router_table,
            key_to_n_atoms_map=key_atom_map))

    def _binary_search_check(
            self, mid_point, routing_table, key_to_n_atoms_map):
        """
        Check function for fix max success.

        :param int mid_point: the point if the list to stop at
        :param ~.UnCompressedMulticastRoutingTable routing_table:
            the basic routing table
        :param dict(int,int) key_to_n_atoms_map:
        :return: true if it compresses
        :rtype: bool
        """

        # convert bitfields into router tables
        bit_field_router_table = self._convert_bitfields_into_router_table(
            routing_table, mid_point, key_to_n_atoms_map)

        # try to compress
        try:
            self._best_routing_entries = self._run_algorithm(
                bit_field_router_table)
            self._best_midpoint = mid_point
            self._compression_attempts[mid_point] = "succcess"
            return True
        except MinimisationFailedError:
            self._compression_attempts[mid_point] = "fail"
            return False
        except PacmanElementAllocationException:
            self._compression_attempts[mid_point] = "Exception"
            return False

    def _run_algorithm(self, router_table):
        """
        Attempts to covert the mega router tables into 1 router table.

        :param list(~.AbstractMulticastRoutingTable) router_table:
            the set of router tables that together need to
            be merged into 1 router table
        :return: compressor router table
        :rtype: list(~.RoutingTableEntry)
        :throws MinimisationFailedError: if it fails to
            compress to the correct length.
        """
        compressor = _PairCompressor(ordered=True)
        compressed_entries = compressor.compress_table(router_table)
        if len(compressed_entries) > Machine.ROUTER_ENTRIES:
            raise MinimisationFailedError(
                f"Compression failed as {len(compressed_entries)} "
                f"entires found")
        return compressed_entries

    def _remove_merged_bitfields_from_cores(self, chip_x, chip_y):
        """
        Goes to SDRAM and removes said merged entries from the cores'
        bitfield region.

        :param int chip_x: the chip x coordinate from which this happened
        :param int chip_y: the chip y coordinate from which this happened
        """
        for entries in self._bit_fields_by_key.values():
            for entry in entries:
                if entry.sort_index < self._best_midpoint:
                    # set merged
                    n_atoms_word = entry.n_atoms_word | self.MERGED_SETTER
                    FecDataView.write_memory(
                        chip_x, chip_y, entry.n_atoms_address, n_atoms_word)

    def _create_table_report(self, router_table, report_out):
        """
        Create the report entry.

        :param ~.AbstractMulticastRoutingTable router_table:
            the uncompressed router table to process
        :param ~io.TextIOBase report_out: the report writer
        """
        n_bit_fields_merged = 0
        n_packets_filtered = 0
        n_possible_bit_fields = 0
        merged_by_core = defaultdict(list)

        for key in self._bit_fields_by_key:
            for bf_data in self._bit_fields_by_key[key]:
                n_possible_bit_fields += 1
                if bf_data.sort_index >= self._best_midpoint:
                    continue
                n_bit_fields_merged += 1
                as_array = bf_data.bit_field_as_bit_array()
                n_packets_filtered += sum(as_array)
                merged_by_core[bf_data.processor_id].append(bf_data)

        percentage_done = 100
        if n_possible_bit_fields != 0:
            percentage_done = (
                (100.0 / float(n_possible_bit_fields)) *
                float(n_bit_fields_merged))

        report_out.write(
            f"\nTable {router_table.x}:{router_table.y} has integrated "
            f"{n_bit_fields_merged} out of {n_possible_bit_fields} available "
            "chip level bitfields into the routing table, thereby producing a "
            f"compression of {percentage_done}%.\n\n")

        report_out.write(
            "The uncompressed routing table had "
            f"{router_table.number_of_entries} entries, the compressed "
            f"one with {n_bit_fields_merged} integrated bitfields has "
            f"{len(self._best_routing_entries)} entries.\n\n")

        report_out.write(
            f"The integration of {n_bit_fields_merged} bitfields removes up "
            f"to {n_packets_filtered} MC packets that otherwise would be "
            "being processed by the cores on the chip, just to be dropped as "
            "they do not target anything.\n\n")

        report_out.write("The compression attempts are as follows:\n\n")
        for mid_point, result in self._compression_attempts.items():
            report_out.write(f"Midpoint {mid_point}: {result}\n")

        report_out.write("\nThe bit_fields merged are as follows:\n\n")

        for core in merged_by_core:
            for bf_data in merged_by_core[core]:
                report_out.write(
                    f"bitfield on core {core} for "
                    f"key {bf_data.master_pop_key}\n")

        report_out.write("\n\n\n")
        report_out.write("The final routing table entries are as follows:\n\n")

        report_out.write(
            "{: <5s} {: <10s} {: <10s} {: <10s} {: <7s} {}\n".format(
                "Index", "Key", "Mask", "Route", "Default", "[Cores][Links]"))
        report_out.write(
            "{:-<5s} {:-<10s} {:-<10s} {:-<10s} {:-<7s} {:-<14s}\n".format(
                "", "", "", "", "", ""))

        entry_count = 0
        n_defaultable = 0
        # Note: _best_routing_table is a list(), router_table is not
        for entry in self._best_routing_entries:
            index = entry_count & self._LOWER_16_BITS
            entry_str = format_route(entry.to_MulticastRoutingEntry())
            entry_count += 1
            if entry.defaultable:
                n_defaultable += 1
            report_out.write(f"{index:>5d} {entry_str}\n")
        report_out.write(f"{n_defaultable} Defaultable entries\n")



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BitFieldSummary(object):
    """
    Summary description of generated bitfields.
    """

    def __init__(
            self, total_merged, max_per_chip, lowest_per_chip, total_to_merge,
            max_to_merge_per_chip, low_to_merge_per_chip,
            average_per_chip_merged, average_per_chip_to_merge):
        """
        :param int total_merged:
        :param int max_per_chip:
        :param int lowest_per_chip:
        :param int total_to_merge:
        :param int max_to_merge_per_chip:
        :param int low_to_merge_per_chip:
        :param float average_per_chip_merged:
        :param float average_per_chip_to_merge:
        """
        self._total_merged = total_merged
        self._max_per_chip = max_per_chip
        self._lowest_per_chip = lowest_per_chip
        self._total_to_merge = total_to_merge
        self._max_to_merge_per_chip = max_to_merge_per_chip
        self._low_to_merge_per_chip = low_to_merge_per_chip
        self._average_per_chip_merged = average_per_chip_merged
        self._average_per_chip_to_merge = average_per_chip_to_merge

    @property
    def total_merged(self):
        """
        :rtype: int
        """
        return self._total_merged

    @property
    def max_per_chip(self):
        """
        :rtype: int
        """
        return self._max_per_chip

    @property
    def lowest_per_chip(self):
        """
        :rtype: int
        """
        return self._lowest_per_chip

    @property
    def total_to_merge(self):
        """
        :rtype: int
        """
        return self._total_to_merge

    @property
    def max_to_merge_per_chip(self):
        """
        :rtype: int
        """
        return self._max_to_merge_per_chip

    @property
    def low_to_merge_per_chip(self):
        """
        :rtype: int
        """
        return self._low_to_merge_per_chip

    @property
    def average_per_chip_merged(self):
        """
        :rtype: float
        """
        return self._average_per_chip_merged

    @property
    def average_per_chip_to_merge(self):
        """
        :rtype: float
        """
        return self._average_per_chip_to_merge



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os.path
from spinn_utilities.log import FormatAdapter
from pacman.model.graphs.application import ApplicationVertex
from spinn_front_end_common.data import FecDataView

logger = FormatAdapter(logging.getLogger(__name__))

_FILENAME = "network_specification.rpt"


def network_specification():
    """
    Generate report on the user's network specification.
    """
    filename = os.path.join(FecDataView.get_run_dir_path(), _FILENAME)
    try:
        with open(filename, "w", encoding="utf-8") as f:
            f.write("*** Vertices:\n")
            for vertex in FecDataView.iterate_vertices():
                _write_report(f, vertex)
    except IOError:
        logger.exception("Generate_placement_reports: Can't open file {}"
                         " for writing.", filename)


def _write_report(f, vertex):
    """
    :param ~io.FileIO f:
    :param ~pacman.model.graphs.application.ApplicationVertex vertex:
    """
    if isinstance(vertex, ApplicationVertex):
        f.write(
            f"Vertex {vertex.label}, size: {vertex.n_atoms}, "
            f"model: {vertex.__class__.__name__}, "
            f"max_atoms: {vertex.get_max_atoms_per_core()}\n")
    else:
        f.write(f"Vertex {vertex.label}, model: {vertex.__class__.__name__}\n")

    if vertex.get_fixed_location():
        f.write(f"    Fixed at:{vertex.get_fixed_location()}\n")

    f.write("    Outgoing Edge Partitions:\n")
    for partition in \
            FecDataView.get_outgoing_edge_partitions_starting_at_vertex(
                vertex):
        f.write(f"    Partition {partition.identifier}:\n")
        for edge in partition.edges:
            f.write(
                f"        Edge: {edge.label}, "
                f"From {edge.pre_vertex.label} to {edge.post_vertex.label}, "
                f"model: {edge.__class__.__name__}\n")
    f.write("\n")



# Copyright (c) 2017 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import logging
import os
import sqlite3
from spinn_utilities.log import FormatAdapter
from spinnman.spalloc.spalloc_job import SpallocJob
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import DsDatabaseException
from spinn_front_end_common.utilities.sqlite_db import SQLiteDB
from .data_row_writer import DataRowWriter

_DDL_FILE = os.path.join(os.path.dirname(__file__), "dse.sql")
logger = FormatAdapter(logging.getLogger(__name__))


class DsSqlliteDatabase(SQLiteDB):
    """
    A database for holding data specification details.
    """
    __slots__ = [
        # The root ethernet id if required
        "__root_ethernet_id"
    ]

    def __init__(self, init_file=None):
        """
        :param bool init_file:
            Whether to initialise the DB from our DDL file. If not specified,
            this is guessed from whether we can read the file.
        """
        database_file = self.default_database_file()

        if init_file is None:
            init_file = not os.path.exists(database_file)

        super().__init__(
            database_file, ddl_file=_DDL_FILE if init_file else None)
        if init_file:
            self.__init_db_contents()
        self.__root_ethernet_id = self.__find_root_id()

    @classmethod
    def default_database_file(cls):
        return os.path.join(FecDataView.get_run_dir_path(),
                            f"ds{FecDataView.get_reset_str()}.sqlite3")

    def __init_db_contents(self):
        """
        Set up the database contents from the machine.
        """
        eth_chips = FecDataView.get_machine().ethernet_connected_chips
        with self.transaction() as cursor:
            cursor.executemany(
                """
                INSERT INTO ethernet(
                    ethernet_x, ethernet_y, ip_address)
                VALUES(?, ?, ?)
                """, (
                    (ethernet.x, ethernet.y, ethernet.ip_address)
                    for ethernet in eth_chips))

    def __find_root_id(self):
        first_x = first_y = root_id = None
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT ethernet_id, ethernet_x, ethernet_y FROM ethernet
                    ORDER BY ethernet_x, ethernet_y
                    LIMIT 1
                    """):
                root_id = row["ethernet_id"]
                first_x = row["ethernet_x"]
                first_y = row["ethernet_y"]
        if root_id is None:
            # Should only be reachable for an empty machine
            raise DsDatabaseException("No ethernet chip found")
        if first_x or first_y:
            logger.warning(
                "No Ethernet chip found at 0,0 using {},{} "
                "for all boards with no IP address.", first_x, first_y)
        return root_id

    def write_data_spec(self, core_x, core_y, core_p, spec_bytes):
        """
        :param int core_x:
            X coordinate of the core that `spec_bytes` applies to
        :param int core_y:
            Y coordinate of the core that `spec_bytes` applies to
        :param int p: Processor ID of the core that `spec_bytes` applies to
        :param bytes spec_bytes: the data specification byte-code
        """
        chip = FecDataView().get_chip_at(core_x, core_y)
        with self.transaction() as cursor:
            cursor.execute(
                """
                INSERT INTO core(
                    x, y, processor, content, ethernet_id, app_id)
                VALUES(?, ?, ?, ?, (
                    SELECT COALESCE((
                        SELECT ethernet_id FROM ethernet
                        WHERE ethernet_x = ? AND ethernet_y = ?
                    ), ?)), ?)
                """, (
                    core_x, core_y, core_p, sqlite3.Binary(spec_bytes),
                    chip.nearest_ethernet_x, chip.nearest_ethernet_y,
                    self.__root_ethernet_id, FecDataView().get_app_id()))

    def get_ds(self, x, y, p):
        """
        Retrieves the data specification as byte-code for this core.

        :param int x: core X coordinate
        :param int y: core Y coordinate
        :param int p: core processor ID
        :return: data specification as byte code
        :rtype: bytes
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT content FROM core
                    WHERE x = ? AND y = ? AND processor = ?
                    LIMIT 1
                    """, (x, y, p)):
                return row["content"]
        return b""

    def keys(self):
        """
        Yields the keys.

        .. note::
            Do not use the database for anything else while iterating.

        :return: Yields the (x, y, p)
        :rtype: iterable(tuple(int,int,int))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT x, y, processor FROM core
                    WHERE content IS NOT NULL
                    """):
                yield (row["x"], row["y"], row["processor"])

    def items(self):
        """
        Yields the keys and values for the data specification data.

        .. note::
            Do not use the database for anything else while iterating.

        :return:
            Yields the (x, y, p) and saved data specification byte-code pairs
        :rtype: iterable(tuple(tuple(int,int,int),~io.RawIOBase))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT x, y, processor, content FROM core
                    WHERE content IS NOT NULL
                    """):
                yield ((row["x"], row["y"], row["processor"]),
                       io.BytesIO(row["content"]))

    def system_items(self):
        """
        Yields the keys and values for the data specification data for system
        cores.

        .. note::
            Do not use the database for anything else while iterating.

        :return: Yields the (x, y, p), saved data specification byte-code, and
            region_size triples
        :rtype: iterable(tuple(tuple(int,int,int),~io.RawIOBase, int))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT x, y, processor, content, memory_used FROM core
                    WHERE content IS NOT NULL AND is_system = 1
                    """):
                yield ((row["x"], row["y"], row["processor"]),
                       io.BytesIO(row["content"]), row["memory_used"])

    def app_items(self):
        """
        Yields the keys and values for the data specification data for
        application cores.

        .. note::
            Do not use the database for anything else while iterating.

        :return:
            Yields the (x, y, p) and saved data specification byte-code pairs
        :rtype: iterable(tuple(tuple(int,int,int),~io.RawIOBase, int))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT x, y, processor, content, memory_used  FROM core
                    WHERE content IS NOT NULL AND is_system = 0
                    """):
                yield ((row["x"], row["y"], row["processor"]),
                       io.BytesIO(row["content"]), row["memory_used"])

    def ds_n_cores(self):
        """
        Returns the number for cores there is a data specification saved for.

        :rtype: int
        :raises DsDatabaseException:
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT COUNT(*) as count FROM core
                    WHERE content IS NOT NULL
                    LIMIT 1
                    """):
                return row["count"]
        raise DsDatabaseException("Count query failed")

    def ds_n_app_cores(self):
        """
        Returns the number for application cores there is a data specification
        saved for.

        :rtype: int
        :raises DsDatabaseException:
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT COUNT(*) as count FROM core
                    WHERE content IS NOT NULL AND is_system = 0
                    LIMIT 1
                    """):
                return row["count"]
        raise DsDatabaseException("Count query failed")

    def ds_n_system_cores(self):
        """
        Returns the number for system cores there is a data specification
        saved for.

        :rtype: int
        :raises DsDatabaseException:
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT COUNT(*) as count FROM core
                    WHERE content IS NOT NULL  AND is_system = 1
                    LIMIT 1
                    """):
                return row["count"]
        raise DsDatabaseException("Count query failed")

    def set_app_id(self, app_id):
        """
        Sets the same app_id for all rows that have data specification content.

        :param int app_id: value to set
        """
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    app_id = ?
                WHERE content IS NOT NULL
                """, (app_id,))

    def get_app_id(self, x, y, p):
        """
        Gets the `app_id` set for this core.

        :param int x: core X coordinate
        :param int y: core Y coordinate
        :param int p: core processor ID
        :rtype: int
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT app_id FROM core
                    WHERE x = ? AND y = ? AND processor = ?
                    LIMIT 1
                    """, (x, y, p)):
                return row["app_id"]
        return None

    def mark_system_cores(self, core_subsets):
        """
        Flags a set of processors as running system binaries.

        :param ~spinn_machine.CoreSubsets core_subsets:
            Which cores to mark.
        """
        with self.transaction() as cursor:
            cursor.executemany(
                """
                UPDATE core SET
                    is_system = 1
                WHERE x = ? AND y = ? AND processor = ?
                """, (
                    (subset.x, subset.y, p)
                    for subset in core_subsets
                    for p in subset.processor_ids))

    def get_write_info(self, x, y, p):
        """
        Gets the provenance returned by the data specification executor.

        :param int x: core X coordinate
        :param int y: core Y coordinate
        :param int p: core processor ID
        :return: start_address, memory_used, memory_written
        :rtype: DataWritten
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT start_address, memory_used, memory_written
                    FROM core
                    WHERE x = ? AND y = ? AND processor = ?
                    LIMIT 1
                    """, (x, y, p)):
                return (row["start_address"], row["memory_used"],
                        row["memory_written"])
        raise ValueError(f"No info for {x}:{y}:{p}")

    def set_write_info(
            self, x, y, p, start, used, written):
        """
        Sets the provenance returned by the data specification executor.

        :param int x: core X coordinate
        :param int y: core Y coordinate
        :param int p: core processor ID
        :param int start: base address
        :param int used: size allocated
        :param int written: bytes written
        """
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    start_address = ?,
                    memory_used = ?,
                    memory_written = ?
                WHERE x = ? AND y = ? AND processor = ?
                """, (start, used, written, x, y, p))
            if cursor.rowcount == 0:
                chip = FecDataView().get_chip_at(x, y)
                cursor.execute(
                    """
                    INSERT INTO core(
                        x, y, processor, start_address,
                        memory_used, memory_written, ethernet_id)
                    VALUES(?, ?, ?, ?, ?, ?, (
                        SELECT COALESCE((
                            SELECT ethernet_id FROM ethernet
                            WHERE ethernet_x = ? AND ethernet_y = ?
                        ), ?)))
                    """, (
                        x, y, p, start, used, written, chip.nearest_ethernet_x,
                        chip.nearest_ethernet_y, self.__root_ethernet_id))

    def set_size_info(self, x, y, p, memory_used):
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    memory_used = ?
                WHERE x = ? AND y = ? AND processor = ?
                """, (memory_used, x, y, p))
            if cursor.rowcount == 0:
                chip = FecDataView().get_chip_at(x, y)
                cursor.execute(
                    """
                    INSERT INTO core(
                        x, y, processor, memory_used, ethernet_id)
                    VALUES(?, ?, ?, ?, (
                        SELECT COALESCE((
                            SELECT ethernet_id FROM ethernet
                            WHERE ethernet_x = ? AND ethernet_y = ?
                        ), ?)))
                    """, (
                        x, y, p, int(memory_used),
                        chip.nearest_ethernet_x, chip.nearest_ethernet_y,
                        self.__root_ethernet_id))

    def clear_write_info(self):
        """
        Clears the provenance for all rows.
        """
        with self.transaction() as cursor:
            cursor.execute(
                """
                UPDATE core SET
                    start_address = NULL,
                    memory_used = NULL,
                    memory_written = NULL
                """)

    def info_n_cores(self):
        """
        Returns the number for cores there is a info saved for.

        :rtype: int
        :raises DsDatabaseException:
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT count(*) as count FROM core
                    WHERE start_address IS NOT NULL
                    LIMIT 1
                    """):
                return row["count"]
        raise DsDatabaseException("Count query failed")

    def info_iteritems(self):
        """
        Yields the keys and values for the Info data.

        .. note::
            A DB transaction may be held while this iterator is processing.
            Reentrant use of this class is not supported.

        :return: Yields the (x, y, p), start_address, memory_used
            and memory_written
        :rtype: iterable(tuple(tuple(int, int, int), int, int, int))
        """
        with self.transaction() as cursor:
            for row in cursor.execute(
                    """
                    SELECT
                        x, y, processor,
                        start_address, memory_used, memory_written
                    FROM core
                    WHERE start_address IS NOT NULL
                    """):
                yield ((row["x"], row["y"], row["processor"]),
                       row["start_address"], row["memory_used"],
                       row["memory_written"])

    def create_data_spec(self, x, y, p):
        """
        Get a stream that will write the data specification for a particular
        core to this database.

        :param int x: X coordinate of the core
        :param int y: Y coordinate of the core
        :param int p: Processor ID of the core
        :return: the stream; the write only happens when the stream is closed
        :rtype: DataRowWriter
        """
        return DataRowWriter(x, y, p, self)

    def write_session_credentials_to_db(self):
        """
        Write Spalloc session credentials to the database, if in use.
        """
        # pylint: disable=protected-access
        if not FecDataView.has_allocation_controller():
            return
        mac = FecDataView.get_allocation_controller()
        if mac.proxying:
            # This is now assumed to be a SpallocJobController;
            # can't check that because of import circularity.
            job = mac._job
            if isinstance(job, SpallocJob):
                with self.transaction() as cur:
                    job._write_session_credentials_to_db(cur)



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import logging

from spinn_utilities.progress_bar import ProgressBar
from spinn_utilities.log import FormatAdapter
from data_specification import DataSpecificationGenerator, ReferenceContext
from data_specification.constants import APP_PTR_TABLE_BYTE_SIZE
from pacman.model.resources import MultiRegionSDRAM, ConstantSDRAM
from spinn_front_end_common.abstract_models import (
    AbstractRewritesDataSpecification, AbstractGeneratesDataSpecification)
from spinn_front_end_common.data import FecDataView
from spinn_front_end_common.utilities.exceptions import ConfigurationException
from spinn_front_end_common.interface.ds import DsSqlliteDatabase
from spinn_front_end_common.utilities.utility_calls import get_report_writer

logger = FormatAdapter(logging.getLogger(__name__))


def graph_data_specification_writer(placement_order=None):
    """
    :param list(~pacman.model.placements.Placement) placement_order:
        the optional order in which placements should be examined
    :rtype: DsSqlliteDatabase
    :raises ConfigurationException:
        If the DSG asks to use more SDRAM than is available.
    """
    writer = _GraphDataSpecificationWriter()
    return writer.run(placement_order)


class _GraphDataSpecificationWriter(object):
    """
    Executes the data specification generation step.
    """

    __slots__ = (
        # Dict of SDRAM usage by chip coordinates
        "_sdram_usage",
        # Dict of list of vertices by chip coordinates
        "_vertices_by_chip")

    def __init__(self):
        self._sdram_usage = defaultdict(lambda: 0)
        self._vertices_by_chip = defaultdict(list)

    def run(self, placement_order=None):
        """
        :param list(~pacman.model.placements.Placement) placement_order:
            the optional order in which placements should be examined
        :return: DSG targets
        :rtype: DsSqlliteDatabase
        :raises ConfigurationException:
            If the DSG asks to use more SDRAM than is available.
        """
        # iterate though vertices and call generate_data_spec for each
        # vertex
        targets = DsSqlliteDatabase()
        targets.write_session_credentials_to_db()

        if placement_order is None:
            placement_order = FecDataView.iterate_placemements()
            n_placements = FecDataView.get_n_placements()
        else:
            n_placements = len(placement_order)

        progress = ProgressBar(n_placements, "Generating data specifications")
        vertices_to_reset = list()

        # Do in a context of global identifiers
        with ReferenceContext():
            for placement in progress.over(placement_order):
                # Try to generate the data spec for the placement
                vertex = placement.vertex
                generated = self.__generate_data_spec_for_vertices(
                    placement, vertex, targets)

                if generated and isinstance(
                        vertex, AbstractRewritesDataSpecification):
                    vertices_to_reset.append(vertex)

                # If the spec wasn't generated directly, and there is an
                # application vertex, try with that
                if not generated and vertex.app_vertex is not None:
                    generated = self.__generate_data_spec_for_vertices(
                        placement, vertex.app_vertex, targets)
                    if generated and isinstance(
                            vertex.app_vertex,
                            AbstractRewritesDataSpecification):
                        vertices_to_reset.append(vertex.app_vertex)

        # Ensure that the vertices know their regions have been reloaded
        for vertex in vertices_to_reset:
            vertex.set_reload_required(False)

        return targets

    def __generate_data_spec_for_vertices(self, pl, vertex, targets):
        """
        :param ~.Placement pl: placement of machine graph to cores
        :param ~.AbstractVertex vertex: the specific vertex to write DSG for.
        :param DsSqlliteDatabase targets:
        :return: True if the vertex was data spec-able, False otherwise
        :rtype: bool
        :raises ConfigurationException: if things don't fit
        """
        # if the vertex can generate a DSG, call it
        if not isinstance(vertex, AbstractGeneratesDataSpecification):
            return False

        with targets.create_data_spec(pl.x, pl.y, pl.p) as data_writer:
            report_writer = get_report_writer(pl.x, pl.y, pl.p)
            spec = DataSpecificationGenerator(data_writer, report_writer)

            # generate the DSG file
            vertex.generate_data_specification(spec, pl)

            # Check the memory usage
            region_size = APP_PTR_TABLE_BYTE_SIZE + sum(spec.region_sizes)

            # extracts the int from the numpy data type generated
            if not isinstance(region_size, int):
                region_size = region_size.item()

            targets.set_size_info(pl.x, pl.y, pl.p, region_size)

            # Check per-region memory usage if possible
            sdram = vertex.sdram_required
            if isinstance(sdram, MultiRegionSDRAM):
                for i, size in enumerate(spec.region_sizes):
                    est_size = sdram.regions.get(i, ConstantSDRAM(0))
                    est_size = est_size.get_total_sdram(
                        FecDataView.get_max_run_time_steps())
                    if size > est_size:
                        # pylint: disable=logging-too-many-args
                        logger.warning(
                            "Region {} of vertex {} is bigger than expected: "
                            "{} estimated vs. {} actual",
                            i, vertex.label, est_size, size)

            self._vertices_by_chip[pl.x, pl.y].append(pl.vertex)
            self._sdram_usage[pl.x, pl.y] += sum(spec.region_sizes)
            if (self._sdram_usage[pl.x, pl.y] <=
                    FecDataView().get_chip_at(pl.x, pl.y).sdram.size):
                return True

        # creating the error message which contains the memory usage of
        # what each core within the chip uses and its original estimate.
        memory_usage = "\n".join(
            "    {}: {} (total={}, estimated={})".format(
                vert, region_size, sum(region_size),
                vert.sdram_required.get_total_sdram(
                    FecDataView.get_max_run_time_steps()))
            for vert in self._vertices_by_chip[pl.x, pl.y])

        raise ConfigurationException(
            f"Too much SDRAM has been used on {pl.x}, {pl.y}.  Vertices and"
            f" their usage on that chip is as follows:\n{memory_usage}")



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct
from spinn_utilities.progress_bar import ProgressBar
from spinnman.messages.sdp import SDPHeader, SDPFlag
from spinnman.messages.scp.abstract_messages import AbstractSCPRequest
from spinnman.messages.scp import SCPRequestHeader
from spinnman.messages.scp.impl import CheckOKResponse
from spinnman.processes import AbstractMultiConnectionProcess
from spinn_front_end_common.utilities.constants import (
    SDP_PORTS, SDP_RUNNING_MESSAGE_CODES)
from spinn_utilities.overrides import overrides


class _UpdateRuntimeRequest(AbstractSCPRequest):
    def __init__(
            self, x, y, p, current_time, run_time, infinite_run, n_sync_steps):
        """
        :param int x:
        :param int y:
        :param int p:
        :param int current_time:
        :param int run_time:
        :param bool infinite_run:
        :param int n_sync_steps:
        """
        # pylint: disable=too-many-arguments
        sdp_flags = SDPFlag.REPLY_EXPECTED

        super().__init__(
            SDPHeader(
                flags=sdp_flags,
                destination_port=SDP_PORTS.RUNNING_COMMAND_SDP_PORT.value,
                destination_cpu=p, destination_chip_x=x, destination_chip_y=y),
            SCPRequestHeader(
                command=SDP_RUNNING_MESSAGE_CODES.SDP_NEW_RUNTIME_ID_CODE),
            argument_1=run_time, argument_2=infinite_run,
            argument_3=current_time,
            data=struct.pack("<I", int(n_sync_steps)))

    @overrides(AbstractSCPRequest.get_scp_response)
    def get_scp_response(self):
        return CheckOKResponse(
            "update runtime",
            SDP_RUNNING_MESSAGE_CODES.SDP_NEW_RUNTIME_ID_CODE.value)


class UpdateRuntimeProcess(AbstractMultiConnectionProcess):
    """
    How to update the target running time of a set of cores.

    .. note::
        The cores must be using the simulation interface.
    """

    def __init__(self, connection_selector):
        """
        :param connection_selector:
        :type connection_selector:
            ~spinnman.processes.abstract_multi_connection_process_connection_selector.AbstractMultiConnectionProcessConnectionSelector
        """
        super().__init__(connection_selector)
        self._progress = None

    def __receive_response(self, _response):
        if self._progress is not None:
            self._progress.update()

    def update_runtime(self, current_time, run_time, infinite_run,
                       core_subsets, n_cores, n_sync_steps):
        """
        :param int current_time:
        :param int run_time:
        :param bool infinite_run:
        :param ~spinn_machine.CoreSubsets core_subsets:
        :param int n_cores: Number of cores being updated
        """
        self._progress = ProgressBar(n_cores, "Updating run time")
        for core_subset in core_subsets:
            for processor_id in core_subset.processor_ids:
                self._send_request(
                    _UpdateRuntimeRequest(
                        core_subset.x, core_subset.y, processor_id,
                        current_time, run_time, infinite_run, n_sync_steps),
                    callback=self.__receive_response)
        self._finish()
        self._progress.end()
        self._progress = None
        self.check_for_error()



# Copyright (c) 2016 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.application import ApplicationVertex


@require_subclass(ApplicationVertex)
class AbstractVertexWithEdgeToDependentVertices(
        object, metaclass=AbstractBase):
    """
    A vertex with a dependent vertices, which should be connected to this
    vertex by an edge directly to each of them.
    """

    __slots__ = ()

    @abstractmethod
    def dependent_vertices(self):
        """
        Return the vertices which this vertex depends upon.

        :rtype: iterable(~pacman.model.graphs.application.ApplicationVertex)
        """

    @abstractmethod
    def edge_partition_identifiers_for_dependent_vertex(self, vertex):
        """
        Return the dependent edge identifiers for a particular dependent
        vertex.

        :param ~pacman.model.graphs.application.ApplicationVertex vertex:
        :rtype: iterable(str)
        """



# Copyright (c) 2019 The University of Manchester
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from spinn_utilities.abstract_base import AbstractBase, abstractmethod
from spinn_utilities.require_subclass import require_subclass
from pacman.model.graphs.machine import MachineVertex


@require_subclass(MachineVertex)
class AbstractSupportsBitFieldGeneration(object, metaclass=AbstractBase):
    """
    Marks a vertex that can provide information about bitfields it wants
    generated on-chip.
    """
    __slots__ = ()

    @abstractmethod
    def bit_field_base_address(self, placement):
        """
        Returns the SDRAM address for the bit field table data.

        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the bitfield address
        :rtype: int
        """

    @abstractmethod
    def bit_field_builder_region(self, placement):
        """
        Returns the SDRAM address for the bit field builder data.

        :param ~pacman.model.placements.Placement placement:
        :return: the SDRAM address for the bitfield builder data
        :rtype: int
        """
